from collections import deque
import json
import os
import queue
import sqlite3
from pathlib import Path
import threading
import time

import requests

from .I18N import _
from .settings import Settings

# Stałe – auth_token (CSRF token) oraz ciasteczka
AUTH_TOKEN = ""  # ustaw właściwą wartość
SHM_SESSION = ""  # ustaw właściwą wartość
SHM_USER = ""  # ustaw właściwą wartość

POST_URL = "http://192.168.1.11:8001/auto_tag/import"
BATCH_SIZE = 1000  # maksymalna liczba wierszy w jednej partii


def flatten_graph(graph):
    """
    Dla każdego tagu (klucza) oblicza domknięcie przechodnie –
    czyli zbiór wszystkich tagów osiągalnych w grafie.
    Używamy cache, aby uniknąć wielokrotnych obliczeń.
    """
    cache = {}

    def dfs(tag):
        if tag in cache:
            return cache[tag]
        result = set()
        # Przechodzimy po bezpośrednich konsekwencjach
        for nxt in graph.get(tag, set()):
            result.add(nxt)
            result |= dfs(nxt)
        cache[tag] = result
        return result

    flattened = {}
    for tag in graph.keys():
        flattened[tag] = dfs(tag)
    return flattened


class TagsRepo:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.db_path = os.path.join(
            os.path.dirname(settings.get_settings_path()), "tags.db"
        )
        regenerate = False
        if not Path(self.db_path).is_file():
            regenerate = True
            print(
                _("Database file not found: {path}, will regenerate DB").format(
                    path=self.db_path
                )
            )
        self.init_tags_db()
        if regenerate:
            self.regenerate_db()

    def get_conn(self):
        return sqlite3.connect(self.db_path)

    # --- Inicjalizacja bazy tagów ---
    def init_tags_db(self):
        try:
            conn = self.get_conn()
            cursor = conn.cursor()
            tables = {
                "tags": """
                    CREATE TABLE IF NOT EXISTS "tags" (
                        "index" INTEGER,
                        "id" INTEGER,
                        "name" TEXT,
                        "post_count" INTEGER,
                        "category" INTEGER,
                        "created_at" TIMESTAMP,
                        "updated_at" TIMESTAMP,
                        "is_deprecated" INTEGER,
                        "words" TEXT
                    )
                """,
                "tag_aliases": """
                    CREATE TABLE IF NOT EXISTS "tag_aliases" (
                        "index" INTEGER,
                        "alias" TEXT,
                        "tag" TEXT
                    )
                """,
                "tag_closure": """
                    CREATE TABLE IF NOT EXISTS "tag_closure" (
                        antecedent TEXT NOT NULL,
                        consequent TEXT NOT NULL,
                        depth INTEGER NOT NULL,
                        PRIMARY KEY (antecedent, consequent)
                    )
                """,
                "tag_implications": """
                    CREATE TABLE IF NOT EXISTS "tag_implications" (
                        antecedent TEXT NOT NULL,
                        consequent TEXT NOT NULL,
                        PRIMARY KEY (antecedent, consequent)
                    )
                """,
            }
            indexes = {
                "tags": [
                    """CREATE INDEX IF NOT EXISTS ix_tags_index ON tags ("index")""",
                    "CREATE INDEX IF NOT EXISTS tags_index_category ON tags (category)",
                    "CREATE INDEX IF NOT EXISTS tags_index_created_at ON tags (created_at)",
                    "CREATE INDEX IF NOT EXISTS tags_index_id ON tags (id)",
                    "CREATE INDEX IF NOT EXISTS tags_index_is_deprecated ON tags (is_deprecated)",
                    "CREATE INDEX IF NOT EXISTS tags_index_name ON tags (name)",
                    "CREATE INDEX IF NOT EXISTS tags_index_post_count ON tags (post_count)",
                    "CREATE INDEX IF NOT EXISTS tags_index_updated_at ON tags (updated_at)",
                ],
                "tag_aliases": [
                    """CREATE INDEX IF NOT EXISTS ix_tag_aliases_index ON tag_aliases ("index")""",
                    "CREATE INDEX IF NOT EXISTS tag_aliases_index_alias ON tag_aliases (alias)",
                    "CREATE INDEX IF NOT EXISTS tag_aliases_index_tag ON tag_aliases (tag)",
                ],
                "tag_closure": [
                    "CREATE INDEX IF NOT EXISTS idx_closure_antecedent ON tag_closure (antecedent)"
                ],
                "tag_implications": [
                    "CREATE INDEX IF NOT EXISTS idx_implications_antecedent ON tag_implications (antecedent)"
                ],
            }
            for table, create_stmt in tables.items():
                cursor.execute(create_stmt)
            for table, index_list in indexes.items():
                for index_stmt in index_list:
                    cursor.execute(index_stmt)
            conn.commit()
            conn.close()
        except Exception as e:
            print(_("Błąd przy inicjalizacji bazy tagów:"), e)

    def regenerate_db(
        self, progress_queue: queue.Queue = None, cancel_event: threading.Event = None
    ):
        """
        Regenerate the database of tags, aliases, and tag implications.

        Optionally updates a progress_queue with status messages and checks
        cancel_event (a threading.Event) to cancel the operation.
        """
        # Connect to SQLite and clear old data
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        if progress_queue:
            progress_queue.put(("label", _("Czyszczenie bazy danych...")))
        cursor.execute("DELETE FROM tags")
        cursor.execute("DELETE FROM tag_aliases")
        conn.commit()

        rate_limit = 10  # requests per second
        min_interval = 1.0 / rate_limit  # minimum seconds between requests

        # -----------------------
        # Fetch Tags
        # -----------------------
        data_list = []
        last_id = 0
        while True:
            if cancel_event and cancel_event.is_set():
                if progress_queue:
                    progress_queue.put(("label", _("Anulowano pobieranie tagów.")))
                conn.close()
                return

            if progress_queue:
                progress_queue.put(
                    (
                        "label",
                        _("Pobieranie tagów (od ID {last_id})...").format(
                            last_id=last_id
                        ),
                    )
                )
            start_time = time.monotonic()
            url = f"https://danbooru.donmai.us/tags.json?limit=1000&page=a{last_id}"
            response = requests.get(url)
            if response.status_code != 200:
                if progress_queue:
                    progress_queue.put(
                        (
                            "label",
                            _(
                                "Błąd przy pobieraniu tagów od ID {last_id}: HTTP {code}"
                            ).format(last_id=last_id, code=response.status_code),
                        )
                    )
                break

            data = response.json()
            if not data:
                break

            last_id = None
            for item in data:
                if cancel_event and cancel_event.is_set():
                    if progress_queue:
                        progress_queue.put(
                            ("label", _("Anulowano przetwarzanie tagów."))
                        )
                    conn.close()
                    return
                tag_id = item.get("id")
                if not last_id:
                    last_id = tag_id
                name = item.get("name")
                post_count = item.get("post_count")
                category = item.get("category")
                created_at = item.get("created_at")
                updated_at = item.get("updated_at")
                is_deprecated = item.get("is_deprecated")
                words = json.dumps(item.get("words"))
                data_list.append(
                    (
                        tag_id,
                        name,
                        post_count,
                        category,
                        created_at,
                        updated_at,
                        is_deprecated,
                        words,
                    )
                )

            if len(data) < 1000:
                break

            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        if progress_queue:
            progress_queue.put(
                ("label", _("Pobrano {count} tagów...").format(count=len(data_list)))
            )

        data_list = sorted(data_list, key=lambda x: x[0])
        data_list = [(idx,) + row for idx, row in enumerate(data_list)]
        cursor.executemany(
            """
            INSERT INTO tags ("index", id, name, post_count, category, created_at, updated_at, is_deprecated, words)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
            data_list,
        )
        conn.commit()
        data_list = []

        # -----------------------
        # Fetch Tag Aliases
        # -----------------------
        last_id = 0
        while True:
            if cancel_event and cancel_event.is_set():
                if progress_queue:
                    progress_queue.put(
                        ("label", _("Anulowano pobieranie aliasów tagów."))
                    )
                conn.close()
                return

            if progress_queue:
                progress_queue.put(
                    (
                        "label",
                        _("Pobieranie aliasów tagów (od ID {last_id})...").format(
                            last_id=last_id
                        ),
                    )
                )
            start_time = time.monotonic()
            url = (
                f"https://danbooru.donmai.us/tag_aliases.json?limit=1000"
                f"&only=id,antecedent_name,consequent_name&search[status]=active&page=a{last_id}"
            )
            response = requests.get(url)
            if response.status_code != 200:
                if progress_queue:
                    progress_queue.put(
                        (
                            "label",
                            _(
                                "Błąd przy pobieraniu aliasów tagów od ID {last_id}: HTTP {code}"
                            ).format(last_id=last_id, code=response.status_code),
                        )
                    )
                break

            data = response.json()
            if not data:
                break

            last_id = None
            for item in data:
                if cancel_event and cancel_event.is_set():
                    if progress_queue:
                        progress_queue.put(
                            ("label", _("Anulowano przetwarzanie aliasów tagów."))
                        )
                    conn.close()
                    return
                tag_id = item.get("id")
                if not last_id:
                    last_id = tag_id
                antecedent = item.get("antecedent_name")
                consequent = item.get("consequent_name")
                data_list.append((antecedent, consequent))

            if len(data) < 1000:
                break

            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        if progress_queue:
            progress_queue.put(
                (
                    "label",
                    _("Pobrano {count} aliasów tagów...").format(count=len(data_list)),
                )
            )

        data_list = sorted(data_list, key=lambda x: x[0])
        data_list = [(idx,) + row for idx, row in enumerate(data_list)]
        cursor.executemany(
            """
            INSERT INTO tag_aliases ("index", alias, tag)
            VALUES (?, ?, ?)
            """,
            data_list,
        )
        conn.commit()
        data_list = []

        # -----------------------
        # Fetch Tag Categories for Implications
        # -----------------------
        cursor.execute("SELECT name FROM tags WHERE category = 4")
        character_tags = {row[0] for row in cursor.fetchall()}
        cursor.execute("SELECT name FROM tags WHERE category = 3")
        copyright_tags = {row[0] for row in cursor.fetchall()}
        cursor.execute("SELECT name FROM tags WHERE category = 5")
        meta_tags = {row[0] for row in cursor.fetchall()}
        cursor.execute("SELECT name FROM tags WHERE category = 1")
        artist_tags = {row[0] for row in cursor.fetchall()}

        cursor.execute("DELETE FROM tag_implications")  # Optional: reset table
        cursor.execute("DELETE FROM tag_closure")  # Optional: reset table
        conn.commit()

        # -----------------------
        # Fetch Tag Implications
        # -----------------------
        tag_dict = {}
        last_id = 0
        while True:
            if cancel_event and cancel_event.is_set():
                if progress_queue:
                    progress_queue.put(
                        ("label", _("Anulowano pobieranie implikacji tagów."))
                    )
                conn.close()
                return

            if progress_queue:
                progress_queue.put(
                    (
                        "label",
                        _("Pobieranie implikacji tagów (od ID {last_id})...").format(
                            last_id=last_id
                        ),
                    )
                )
            start_time = time.monotonic()
            url = f"https://danbooru.donmai.us/tag_implications.json?limit=1000&page=a{last_id}"
            response = requests.get(url)
            if response.status_code != 200:
                if progress_queue:
                    progress_queue.put(
                        (
                            "label",
                            _(
                                "Błąd przy pobieraniu implikacji tagów od ID {last_id}: HTTP {code}"
                            ).format(last_id=last_id, code=response.status_code),
                        )
                    )
                break

            data = response.json()
            if not data:
                break

            last_id = None
            for item in data:
                if cancel_event and cancel_event.is_set():
                    if progress_queue:
                        progress_queue.put(
                            ("label", _("Anulowano przetwarzanie implikacji tagów."))
                        )
                    conn.close()
                    return
                tag_id = item.get("id")
                if not last_id:
                    last_id = tag_id
                if item.get("status") != "active":
                    continue

                antecedent = item.get("antecedent_name")
                consequent = item.get("consequent_name")

                # Prefix tags based on their category
                if antecedent in character_tags:
                    antecedent = f"character:{antecedent}"
                elif antecedent in copyright_tags:
                    antecedent = f"copyright:{antecedent}"
                elif antecedent in meta_tags:
                    antecedent = f"meta:{antecedent}"
                elif antecedent in artist_tags:
                    antecedent = f"artist:{antecedent}"

                if consequent in character_tags:
                    consequent = f"character:{consequent}"
                elif consequent in copyright_tags:
                    consequent = f"copyright:{consequent}"
                elif consequent in meta_tags:
                    consequent = f"meta:{consequent}"
                elif consequent in artist_tags:
                    consequent = f"artist:{consequent}"

                if antecedent not in tag_dict:
                    tag_dict[antecedent] = set()
                tag_dict[antecedent].add(consequent)

            if len(data) < 1000:
                break

            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        if progress_queue:
            progress_queue.put(
                (
                    "label",
                    _("Pobrano implikacje dla {count} tagów...").format(
                        count=len(tag_dict)
                    ),
                )
            )

        # -----------------------
        # Insert Tag Implications and Build Transitive Closure
        # -----------------------
        for antecedent, consequents in tag_dict.items():
            for consequent in consequents:
                if antecedent != consequent:
                    cursor.execute(
                        "INSERT OR IGNORE INTO tag_implications VALUES (?, ?)",
                        (antecedent, consequent),
                    )
        conn.commit()

        closure_data = self.build_transitive_closure(tag_dict)
        cursor.executemany(
            "INSERT INTO tag_closure VALUES (?, ?, ?)",
            closure_data,
        )
        conn.commit()
        conn.close()

        if progress_queue:
            progress_queue.put(("label", _("Regeneracja bazy zakończona.")))

    def build_transitive_closure(self, tag_dict):
        closure = set()
        for antecedent in tag_dict:
            visited = set()
            queue = deque([(antecedent, 0)])

            while queue:
                current_tag, depth = queue.popleft()
                if current_tag in visited:
                    continue
                visited.add(current_tag)

                # Add to closure if not self-reference
                if current_tag != antecedent:
                    closure.add((antecedent, current_tag, depth))

                # Traverse next level
                for next_tag in tag_dict.get(current_tag, []):
                    queue.append((next_tag, depth + 1))

        return closure


# def garbage():
#
#    # Spłaszczenie struktury – obliczenie domknięcia przechodniego
#    flattened_tag_dict = flatten_graph(tag_dict)
#    print("Spłaszczono strukturę implikacji.")
#
#    # Przygotowanie listy wierszy do CSV
#    # Każdy wiersz: (antecedent, consequents jako space-separated string)
#    csv_rows = []
#    for antecedent, consequents in flattened_tag_dict.items():
#        # Sortujemy, żeby wynik był deterministyczny
#        consequents_str = " ".join(sorted(consequents))
#        csv_rows.append((antecedent, consequents_str))
#
#    print(f"Łącznie wierszy do wysłania: {len(csv_rows)}")
#
#    # Konfiguracja ciasteczek do żądania POST
#    cookies = {
#        'shm_session': SHM_SESSION,
#        'shm_user': SHM_USER
#    }
#
#    # Wysyłanie danych w partiach po BATCH_SIZE wierszy
#    for i in range(0, len(csv_rows), BATCH_SIZE):
#        batch = csv_rows[i:i+BATCH_SIZE]
#        # Utworzenie pliku CSV w pamięci
#        output = io.StringIO()
#        writer = csv.writer(output, quoting=csv.QUOTE_ALL)
#        for row in batch:
#            writer.writerow(row)
#        csv_content = output.getvalue()
#        output.close()
#
#        # Przygotowanie danych formularza (z auth_token) oraz pliku CSV
#        data = {
#            'auth_token': AUTH_TOKEN
#        }
#        files = {
#            'auto_tag_file': ('batch.csv', csv_content, 'text/csv')
#        }
#
#        print(f"Wysyłanie batcha wierszy {i+1} - {i+len(batch)}...")
#        post_response = requests.post(POST_URL, data=data, files=files, cookies=cookies, allow_redirects=False)
#        if post_response.status_code in (200, 302):
#            print(f"Batch {i+1}-{i+len(batch)} wysłany pomyślnie.")
#        else:
#            print(f"Błąd przy wysyłaniu batcha {i+1}-{i+len(batch)}: HTTP {post_response.status_code}")
#
#
#    print("Wszystkie dane zostały wysłane.")
