#!/usr/bin/env python3
import os
import time
import requests
import json
import tempfile
from PIL import Image
from wdtagger import Tagger

# Stałe (ustaw odpowiednie wartości)
LOGIN = ""
API_KEY = ""
AUTH_TOKEN = ""
SHM_SESSION = ""
SHM_USER = ""

TAGGER = Tagger()

# Bazowy URL Kapitanbooru
BASE_URL = "http://192.168.1.11:8001"


# Funkcja wyciągająca tytuł – czyli nazwę pliku bez rozszerzenia
def extract_title(file_name):
    return os.path.splitext(file_name)[0]


# Funkcja aktualizująca rating i tagi dla danego obrazu na serwerze
def update_post(post, new_rating):
    """
    Aktualizuje post na serwerze, ustawiając rating.
    new_rating – rating ustalony przez Taggera (bez mapowania general i sensitive).
    Rating wysyłamy jako pierwsza litera (mała), a kod 302 traktujemy jako sukces.
    """
    post_id = post.get("id")
    title = extract_title(post.get("file_name", ""))
    owner = LOGIN  # Stałe LOGIN
    source = post.get("source", "")
    rating_param = new_rating[0] if new_rating else "?"
    if rating_param == post.get("rating"):
        print(f"Post {post_id} już ma rating {new_rating}, pomijam.")
        return False
    tags = post.get("tag_string")

    url = BASE_URL.rstrip("/") + "/post/set"
    cookies = {"shm_session": SHM_SESSION, "shm_user": SHM_USER}
    data = {
        "auth_token": AUTH_TOKEN,
        "image_id": post_id,
        "title": title,
        "owner": owner,
        "tags": tags,
        "source": source,
        "rating": rating_param,
    }
    try:
        # Ustawiamy allow_redirects=False, aby 302 nie było traktowane jako błąd
        r = requests.post(url, data=data, cookies=cookies, allow_redirects=False)
        if r.status_code in (200, 201, 302):
            print(f"Post {post_id} zaktualizowany, rating: {new_rating}")
        else:
            print(f"Błąd aktualizacji postu {post_id}: {r.status_code} {r.text}")
    except Exception as e:
        print(f"Błąd przy aktualizacji postu {post_id}: {e}")
    return True


def main():
    page = 1
    posts = json.loads("[]")
    total_processed = 0
    while True:
        # Tworzymy URL do pobierania postów
        posts_url = (
            f"{BASE_URL}/posts.json?&tags=rating:s&limit=100&page={page}"
            f"&login={LOGIN}&api_key={API_KEY}"
        )
        try:
            response = requests.get(posts_url)
            if response.status_code != 200:
                print(
                    f"Błąd pobierania posts.json: {response.status_code} {response.text}"
                )
                break
            response_posts = response.json()
        except Exception as e:
            print("Błąd przy pobieraniu JSON:", e)
            break

        if not response_posts:
            print("Brak więcej postów.")
            break

        posts.extend(response_posts)
        print(
            f"Pobrano stronę {page} z {len(response_posts)} postami. Zebrano łącznie {len(posts)} postów."
        )

        # Opcjonalnie: odczekaj chwilę między postami, aby nie przeciążyć serwera
        time.sleep(0.5)
        # Jeśli mniej niż 100 postów na stronie, kończymy
        if len(response_posts) < 100:
            break
        page += 1

    for post in posts:
        total_processed += 1
        print(f"\nPrzetwarzam post {post.get('id')} ({total_processed})...")
        file_url = post.get("file_url")
        if not file_url:
            print("Brak file_url, pomijam.")
            continue
        # Pobieramy obrazek do tymczasowego pliku
        try:
            r = requests.get(file_url, stream=True)
            if r.status_code != 200:
                print(f"Błąd pobierania obrazu: {r.status_code}")
                continue
            with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as tmp_file:
                for chunk in r.iter_content(chunk_size=8192):
                    tmp_file.write(chunk)
                tmp_file_path = tmp_file.name
            print(f"Obrazek zapisany tymczasowo: {tmp_file_path}")
        except Exception as e:
            print("Błąd przy pobieraniu obrazu:", e)
            continue

        # Otwieramy obrazek i uruchamiamy Taggera
        try:
            img = Image.open(tmp_file_path)
        except Exception as e:
            print("Błąd przy otwieraniu obrazu:", e)
            os.remove(tmp_file_path)
            continue

        try:
            result = TAGGER.tag(img)
            new_rating = (
                result.rating
                if result.rating in ["general", "sensitive", "questionable", "explicit"]
                else ""
            )
            print(f"Tagger: rating = {result.rating}")
        except Exception as e:
            print("Błąd Taggera:", e)
            os.remove(tmp_file_path)
            continue
        finally:
            img.close()

        # Aktualizujemy post na serwerze
        updated = update_post(post, new_rating)

        # Usuwamy tymczasowy plik
        try:
            os.remove(tmp_file_path)
            print(f"Tymczasowy plik {tmp_file_path} usunięty.")
        except Exception as e:
            print("Błąd przy usuwaniu tymczasowego pliku:", e)

        # Odczekaj chwilę między postami, aby nie przeciążyć serwera, jeśli aktualizowano
        if updated:
            time.sleep(0.5)

    print(f"\nZakończono przetwarzanie. Łącznie przetworzono {total_processed} postów.")


if __name__ == "__main__":
    main()
