import os
import pickle
import sqlite3
import time
from .I18N import _
from .settings import Settings


class TaggerCache:
    def __init__(self, settings: Settings, tagger_name: str, tagger_version: str):
        self.tagger_name = tagger_name
        self.tagger_version = tagger_version
        self.settings = settings
        self.cache_db_path = os.path.join(
            os.path.dirname(settings.get_settings_path()), "tagger_cache.db"
        )
        self.init_cache_db()
        self.clear_expired_cache()

    # --- Inicjalizacja bazy cache ---
    def init_cache_db(self):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS tagger_cache (
                    file_md5 TEXT PRIMARY KEY,
                    tagger_name TEXT,
                    tagger_version TEXT,
                    created_at INTEGER,
                    result BLOB
                )
            """
            )
            conn.commit()
            conn.close()
        except Exception as e:
            print("Błąd przy inicjalizacji bazy cache:", e)

    def __getitem__(self, file_md5):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            cursor.execute(
                "SELECT tagger_name, tagger_version, created_at, result FROM tagger_cache WHERE file_md5 = ?",
                (file_md5,),
            )
            row = cursor.fetchone()
            conn.close()
            if row:
                created_at = row[2]
                if time.time() - created_at < self.settings.cache_expiry:
                    return {
                        "tagger_name": row[0],
                        "tagger_version": row[1],
                        "created_at": created_at,
                        "result": pickle.loads(row[3]),
                    }
                else:
                    self.delete_cache_entry(file_md5)
        except Exception as e:
            print(_("Błąd przy odczycie cache dla"), file_md5, ":", e)
        return None

    def __setitem__(self, file_md5, result):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            created_at = int(time.time())
            data_blob = sqlite3.Binary(pickle.dumps(result))
            cursor.execute(
                """
                INSERT OR REPLACE INTO tagger_cache (file_md5, tagger_name, tagger_version, created_at, result)
                VALUES (?, ?, ?, ?, ?)
            """,
                (
                    file_md5,
                    self.tagger_name,
                    self.tagger_version,
                    created_at,
                    data_blob,
                ),
            )
            conn.commit()
            conn.close()
        except Exception as e:
            print(_("Błąd przy zapisie cache dla"), file_md5, ":", e)

    def delete_cache_entry(self, file_md5):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            cursor.execute("DELETE FROM tagger_cache WHERE file_md5 = ?", (file_md5,))
            conn.commit()
            conn.close()
        except Exception as e:
            print(_("Błąd przy usuwaniu cache dla"), file_md5, ":", e)

    def clear_expired_cache(self):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            expiry_threshold = int(time.time() - self.settings.cache_expiry)
            cursor.execute(
                "DELETE FROM tagger_cache WHERE created_at < ?", (expiry_threshold,)
            )
            conn.commit()
            conn.close()
        except Exception as e:
            print(_("Błąd przy czyszczeniu przeterminowanego cache:"), e)

    def clear_cache(self):
        try:
            conn = sqlite3.connect(self.cache_db_path)
            cursor = conn.cursor()
            cursor.execute("DELETE FROM tagger_cache")
            conn.commit()
            conn.close()
            return True, None
        except Exception as e:
            return False, e
